from __future__ import annotations
from typing import (
    NamedTuple,
    TypeVar,
    Iterator,
    Iterable,
    Callable,
    Literal,
    Any,
    overload,
)
from . import _oxide
from ._oxide import permanova
from .types import AnySequence
from .distance import (
    get_distance_matrix,
    get_distance_matrix_numba,
    get_distance_matrix_parallel,
)
import numpy as np
import warnings

T = TypeVar("T")
Tc = TypeVar("Tc", covariant=True)


def ordinal_encoding(
    arr: np.ndarray[Any, np.dtype[_oxide.ordinal_encoding_dtypes]],
) -> np.ndarray[Any, np.dtype[np.uint]]:
    """
    An implementation of normal ordinal encoding in rust to shed some dependencies.
    Only works with certain types!
    """
    if not isinstance(arr, np.ndarray):
        raise TypeError("input should be a `np.ndarray`")
    if arr.dtype.name.startswith(("str", "bytes", "void")):
        suffix = "".join(x for x in arr.dtype.name if x.isalpha())
    else:
        suffix = arr.dtype.name
    try:
        func = getattr(_oxide, f"ordinal_encoding_{suffix}")
    except NameError:
        raise ValueError(f"input dtype {arr.dtype} not understood")
    return func(arr)


class PermanovaResults(NamedTuple):
    statistic: float
    pvalue: float


@overload
def calculate_distances(
    things: Iterable[T],
    distance: Callable[[T, T], np.float64],
    engine: Literal["python"],
    symmetrification: Literal["roundtrip", "one-sided"],
) -> np.ndarray[Any, np.dtype[np.floating[Any]]]:
    ...


@overload
def calculate_distances(
    things: AnySequence[T],
    distance: Callable[[T, T], np.float64],
    engine: Literal["numba"],
    symmetrification: Literal["roundtrip", "one-sided"],
) -> np.ndarray[Any, np.dtype[np.floating[Any]]]:
    ...


@overload
def calculate_distances(
    things: AnySequence[T],
    distance: Callable[[T, T], np.float64],
    engine: Literal["concurrent.futures",],
    symmetrification: Literal["roundtrip", "one-sided"],
    workers: int,
) -> np.ndarray[Any, np.dtype[np.floating[Any]]]:
    ...


def calculate_distances(
    things: Iterable[T] | AnySequence[T],
    distance: Callable[[T, T], np.float64],
    engine: Literal["python", "numba", "concurrent.futures"],
    symmetrification: Literal["roundtrip", "one-sided"],
    workers: int | None = None,
) -> np.ndarray[Any, np.dtype[np.floating[Any]]]:
    """
    Compute distances between `things`. Primarily used in `peroxymanova.run`
    but exposed here for edge cases when `run` is inconvenient. For parameter
    explanation see `peroxymanova.run` docs.
    """
    result = _calculate_distances(
        things=things,
        distance=distance,
        engine=engine,
        symmetrification=symmetrification,
        workers=workers,
    )
    if np.isnan(result).any():
        warnings.warn(
            "distances contain `NaN`s, consider removing problematic entities from the list of `things`"
        )
    return result


def _calculate_distances(
    things: Iterable[T] | AnySequence[T],
    distance: Callable[[T, T], np.float64],
    engine: Literal["python", "numba", "concurrent.futures"],
    symmetrification: Literal["roundtrip", "one-sided"],
    workers: int | None = None,
) -> np.ndarray[Any, np.dtype[np.floating[Any]]]:
    if isinstance(things, Iterator):
        raise TypeError(
            "`things` should be immutable on read, i.e. be an `Iterable` and not `Iterator`"
        )
    if workers is not None and engine in ["python", "numba"]:
        raise TypeError("`workers` is an option for engine='concurrent.futures'")

    if not callable(distance):
        raise TypeError("distance must be a callable")

    if engine == "python":
        if not isinstance(things, Iterable):
            raise TypeError("`things` must be a `Iterable` for engine == 'python'")
        return get_distance_matrix(things, distance, symmetrification)

    elif engine == "concurrent.futures":
        if not isinstance(things, AnySequence):
            raise TypeError(
                "`things` must be a `Sequence` for engine == 'concurrent.futures'"
            )
        return get_distance_matrix_parallel(things, distance, symmetrification, workers)

    elif engine == "numba":
        import numba

        return numba.njit(get_distance_matrix_numba)(
            things,
            numba.njit(distance),  # type: ignore
            symmetrification,
        )  # type: ignore

    raise ValueError("engine isnt in the list of allowed ones, consult the type hints")


@overload
def permanova_pipeline(
    things: Iterable[T],
    distance: Callable[[T, T], np.float64],
    labels: np.ndarray[Any, np.dtype[_oxide.ordinal_encoding_dtypes]],
    engine: Literal["python"],
    symmetrification: Literal["roundtrip", "one-sided"],
    already_squared=False,
    permutations=1000,
) -> PermanovaResults:
    ...


@overload
def permanova_pipeline(
    things: AnySequence[T],
    distance: Callable[[T, T], np.float64],
    labels: np.ndarray[Any, np.dtype[_oxide.ordinal_encoding_dtypes]],
    engine: Literal["numba"],
    symmetrification: Literal["roundtrip", "one-sided"],
    already_squared=False,
    permutations=1000,
) -> PermanovaResults:
    ...


@overload
def permanova_pipeline(
    things: AnySequence[T],
    distance: Callable[[T, T], np.float64],
    labels: np.ndarray[Any, np.dtype[_oxide.ordinal_encoding_dtypes]],
    engine: Literal["concurrent.futures"],
    symmetrification: Literal["roundtrip", "one-sided"],
    already_squared=False,
    permutations=1000,
    workers: int | None = None,
) -> PermanovaResults:
    ...


def permanova_pipeline(
    things: Iterable[T] | AnySequence[T],
    distance: Callable[[T, T], np.float64],
    labels: np.ndarray[Any, np.dtype[_oxide.ordinal_encoding_dtypes]],
    engine: Literal["python", "numba", "concurrent.futures"],
    symmetrification: Literal["roundtrip", "one-sided"],
    already_squared=False,
    permutations=1000,
    workers: int | None = None,
) -> PermanovaResults:
    """
    ### Run the full pipeline:
    1. Compute pairwise distances between `things` and resolve inconsistencies*
    2. Perform ordinal encoding of lables for the `peroxymanova.permanova`
    3. Run highly optimized compiled `peroxymanova.permanova`*

    ### Notes
    - "inconsistencies" in the pairwise distances refer to the fact that the
    properties of symmetry (i.e. `distance(a, b) == distance(b, a)`)
    and zero distance to self (i.e. `distance(a, a) == 0`) dont necessarily hold
    an arbitrary user-defined function `distance(a, b)`, and those are important for metrics.
    While violating zero distance to self happens to not affect one-way permutational anova,
    violating symmetry does affect it, so `symmetrification` parameter controls how it is achieved.
    - the result of running permanova is the p-value for the
    hypothesis that different groups of `things` (encoded in `labels`)
    are identical, as far as the `distance` metric is concerned.
    Hence, it can be interpreted in the same way a p-value for anova is interpreted.
    - The key difference between permanova and anova is that this algorithm only requires
    a `distance` to be defined, while anova needs all operators on numbers (like addition
    and division and whatnot)

    ### Parameters
    for the types please consult the type annotations

    things: a set of (any) things of some type `T` to run permanova for

    distance: a function that accepts two objects of type `T` and returns\
    a float "distance" between them. This can be anything, really,\
    but the closer it is in spirit to the Eucledian (L2) distance, the better\
    (but then again, if an honest Eucledian distance can actually be defined,\
    then `T` is just some kind of number and you should use anova from scipy)

    labels: an array of labels of the same length as `things`\
    (even if `__len__` is not defined for `things`, there must be a correspondence by order)

    engine: the engine that will calculate the distance matrix:
    - python: the most flexible one, only requires `things` to have `__iter__` method.\
    It is implied that iterating over `things` doesnt mutate them.
    - numba: uses numba's just-in-time compilation to calculate the distance matrix faster,\
    but requires `things` to have `__getitem__` method and be numba-friendly,\
    along with the `distance` function. If you arent sure if your objects are numba-friendly,\
    prepare for numba errors.
    - concurrent.futures: uses concurrent.futures to run the distance computation in parallel,\
    requires `things` to have `__getitem__` method. This can be used for relatively fast computation\
    of potentially large objects, as the `things` can be a lazy dataset that loads huge things\
    with its `__getitem__`.

    symmetrification: a strategy for ensuring symmetric distance matrix
    - roundtrip: each `distance(a, b)` is summed with its counterpart `distance(b, a)`
    - one-sided: only compute one `distance(a, b)`\
    for `a` and `b` such that `a` comes before `b` in the `things`

    already_squared: permanova algorithm requires distances to be squared,\
    but for speed one can provide a `distance` function that returns a squared result\
    (it makes sense if `distance` was supposed to take a square root)

    permutations: amount of permutations for calculating p-value in the core algorithm

    workers: amount of workers for concurrent.futures, only makes sense for that engine

    ### Returns
    PermanovaResults(statistic: np.float64, pvalue: np.float64)
    `statistic` is a deterministic number that represents how
    extreme the difference between groups is. It is kind of impossible to interpret
    but its useful to compare different runs on the same data.
    `pvalue` is a permutation-based approximation of the
    probability that the null hypothesis is true. Since the null hypothesis here
    is that there is no difference between groups, a low p-value (typically below 0.05)
    means that groups are likely different.
    """
    if (
        labels.dtype.kind in ["i", "u"]
        and min(labels) == 0
        and max(labels) == len(np.unique(labels)) - 1
    ):
        fastlabels = labels.astype(np.uint, copy=False)  # if possible, dont copy
    else:
        fastlabels = ordinal_encoding(labels)
    dist = _calculate_distances(things, distance, engine, symmetrification, workers)
    if not already_squared:
        dist **= 2
    return PermanovaResults(*permanova(dist, fastlabels, permutations))



from reference import square, packed
import peroxymanova
import cProfile
import numpy as np
from typing import Callable, Any
import pstats
import inspect
import re


class Color:
    reset = "\u001b[0m"

    def __init__(self, value) -> None:
        self.value = value

    def __add__(self, other: str):
        if not isinstance(other, str):
            return NotImplemented
        return self.value + other + self.reset


YELLOW = Color("\033[1;33m")

size = 1000
dist = np.random.random((size, size))
dist = dist + dist.T
np.fill_diagonal(dist, 0)
labels = np.random.randint(0, 3, size)


def run(callback: Callable[[], Any]):
    code = re.search(r"lambda: (.+)", inspect.getsource(callback))
    if code is None:
        raise ValueError("a lambda should be passed")
    print(YELLOW + code.group(1), "\n")
    with cProfile.Profile() as pr:
        callback()
    pstats.Stats(pr).sort_stats("tottime").print_stats(5)


run(lambda: square.permanova(dist, labels))
run(lambda: packed.permanova(dist, labels))
run(lambda: peroxymanova.permanova(dist, labels.astype("uint")))



import importlib.metadata
from ._oxide import permanova
from ._peroxymanova import (
    calculate_distances,
    permanova_pipeline,
    ordinal_encoding,
    PermanovaResults,
)

__all__ = [
    "permanova",
    "calculate_distances",
    "permanova_pipeline",
    "ordinal_encoding",
    "PermanovaResults",
]

if __package__ is not None:
    __version__ = importlib.metadata.version(__package__)



from __future__ import annotations
from typing import Any, Generic, TypeVar
import peroxymanova
import numpy as np
from scipy.spatial import distance_matrix
from scipy.stats import f_oneway
import pytest

size = 100
objects = np.random.random((size, 1))
dist: np.ndarray[Any, np.dtype[np.float64]] = distance_matrix(objects, objects)
labels = np.random.randint(0, 2, size)
anova = f_oneway(*(objects[labels == i] for i in np.unique(labels)))

T = TypeVar("T", covariant=True)


def distance_function(a: np.float64, b: np.float64):
    return np.float64(np.sqrt(np.sum((a - b) ** 2)))


class MockDataLoaderIndexable(Generic[T]):
    def __init__(self, arr: np.ndarray) -> None:
        self.arr = arr

    def __getitem__(self, __key: int) -> T:
        return self.arr[__key]

    def __len__(self) -> int:
        return len(self.arr)


def test_permanova():
    our = peroxymanova.permanova(dist**2, labels.astype(np.uint))
    assert np.allclose(anova.statistic[0], our[0])


@pytest.mark.parametrize("labeltype", [np.int16, np.int64, np.str_])
@pytest.mark.parametrize("engine", ["python", "numba"])
@pytest.mark.parametrize("symmetrification", ["roundtrip", "one-sided"])
def test_pipeline_run_iterable(labeltype, engine, symmetrification):
    run_py_results = peroxymanova.permanova_pipeline(
        objects,
        distance_function,
        labels.astype(labeltype),
        engine=engine,
        symmetrification=symmetrification,
    )
    assert np.allclose(anova.statistic[0], run_py_results.statistic)


@pytest.mark.parametrize("labeltype", [np.int16, np.int64])
def test_pipeline_run_ordinal_encoding_on_ints(labeltype):
    run_py_results = peroxymanova.permanova_pipeline(
        objects,
        distance_function,
        labels.astype(labeltype) + 42,
        engine="python",
        symmetrification="roundtrip",
    )
    assert np.allclose(anova.statistic[0], run_py_results.statistic)


@pytest.mark.parametrize("labeltype", [np.int16, np.int64, np.str_])
@pytest.mark.parametrize("symmetrification", ["roundtrip", "one-sided"])
def test_pipeline_run_indexable(labeltype, symmetrification):
    run_py_results = peroxymanova.permanova_pipeline(
        MockDataLoaderIndexable(objects),
        distance_function,
        labels.astype(labeltype),
        engine="concurrent.futures",
        symmetrification=symmetrification,
    )
    assert np.allclose(anova.statistic[0], run_py_results.statistic)


@pytest.mark.parametrize("engine", ["python", "numba"])
def test_pipeline_clean_stdout(engine, capsys: pytest.CaptureFixture[str]):
    peroxymanova.permanova_pipeline(
        objects,
        distance_function,
        labels,
        engine=engine,
        symmetrification="roundtrip",
    )
    assert not capsys.readouterr().out


@pytest.mark.parametrize("engine", ["python", "numba"])
def test_pipeline_original_data_intact(engine):
    previousdist = dist.copy()
    previouslables = labels.copy()
    peroxymanova.permanova_pipeline(
        objects,
        distance_function,
        labels,
        engine=engine,
        symmetrification="roundtrip",
    )
    assert np.allclose(previousdist, dist)
    assert np.allclose(previouslables, labels)


@pytest.mark.parametrize("ngroups", range(2, 10))
def test_permanova_different_number_of_groups(ngroups):
    labels = np.random.randint(0, ngroups, size)
    our = peroxymanova.permanova(dist**2, labels.astype(np.uint))
    anova = f_oneway(*(objects[labels == i] for i in np.unique(labels)))
    assert np.allclose(our[0], anova.statistic[0])


@pytest.mark.parametrize("npermutations", range(3))
def test_permanova_different_number_of_permutations(npermutations):
    labels = np.random.randint(0, 5, size)
    our = peroxymanova.permanova(
        dist**2, labels.astype(np.uint), permutations=npermutations * 10
    )
    anova = f_oneway(*(objects[labels == i] for i in np.unique(labels)))
    assert np.allclose(our[0], anova.statistic[0])


class TestGracefulValueErrors:
    def test_permanova_dist(self):
        with pytest.raises(ValueError):
            print(
                peroxymanova.permanova(
                    np.random.random((size, size + 1)), labels.astype("uint")
                )
            )

    def test_permanova_non_onehot_labels(self):
        with pytest.raises(ValueError):
            print(peroxymanova.permanova(dist, (labels + 42).astype("uint")))

    def test_permanova_empty_labels(self):
        with pytest.raises(ValueError):
            print(peroxymanova.permanova(dist, np.zeros(0).astype("uint")))

    def test_permanova_empty_both(self):
        with pytest.raises(ValueError):
            print(peroxymanova.permanova(np.zeros((0, 0)), np.zeros(0).astype("uint")))

    def test_permanova_empty_dist(self):
        with pytest.raises(ValueError):
            print(
                peroxymanova.permanova(np.random.random((0, 0)), labels.astype("uint"))
            )

    def test_permanova_one_class(self):
        with pytest.raises(ValueError):
            print(peroxymanova.permanova(dist, np.zeros(size).astype("uint")))


def test_graceful_warning_nan():
    with pytest.warns():
        print(
            peroxymanova.calculate_distances(
                [np.nan] * 100, lambda a, b: np.float64(a - b), "python", "roundtrip"
            )
        )


type_errors_par = pytest.mark.parametrize(
    "sillytype", [None, object, Generic, lambda: 42]
)


class TestGracefulTypeErrors:
    @type_errors_par
    def test_permanova_dist(self, sillytype):
        with pytest.raises(TypeError):
            print(f"{sillytype = }")
            peroxymanova.permanova(sillytype, labels.astype("uint"))

    @type_errors_par
    def test_permanova_labels(self, sillytype):
        with pytest.raises(TypeError):
            print(f"{sillytype = }")
            peroxymanova.permanova(dist, sillytype)

    @type_errors_par
    def test_ordinal(self, sillytype):
        with pytest.raises(TypeError):
            print(f"{sillytype = }")
            peroxymanova.ordinal_encoding(sillytype)

    @type_errors_par
    def test_calculate_distances_things(self, sillytype):
        with pytest.raises(TypeError):
            print(f"{sillytype = }")
            peroxymanova.calculate_distances(
                sillytype, lambda a, b: np.float64(a - b), "python", "roundtrip"
            )

    @type_errors_par
    def test_calculate_distances_lambda(self, sillytype):
        with pytest.raises(TypeError):
            print(f"{sillytype = }")
            peroxymanova.calculate_distances(
                objects, lambda a, b: sillytype, "python", "roundtrip"
            )


@pytest.mark.parametrize("dtype", ["str", "int64", "int32", "int16"])
def test_ordinal_encoding_dtypes(dtype):
    labels = np.random.randint(0, 15, 1000)
    res = peroxymanova.ordinal_encoding(labels.astype(dtype))
    assert res.min() == 0 and res.max() == 14



import numpy as np
from typing import Any

def permanova(
    sqdistances: np.ndarray[Any, np.dtype[np.float64]],
    labels: np.ndarray[Any, np.dtype[np.uint]],
    permutations=1000,
) -> tuple[float, float]:
    """
    ### Notes
    Run a highly optimized permanova implementation written in rust and compiled for your architecture.
    The question the algorithm answers is "are these observations from different
    groups coming from the same distribution?", or, in simpler terms,
    "is there a real difference between these groups?"
    This algorithm relies on the fact that the sum of squared distances to
    a centroid of a group of measurements equals the average distance between
    every measurement in that group. To see the full explanation refer to the original paper.

    ### Parameters

    sqdistances: a 2d distance matrix that is already squared (as in, literally, x**2 kind of squared)

    labels: a 1d array of uint category labels, where a given `labels[i]`\
        refers to a category label of `sqdistances[i,:]` and `sqdistances[:, i]`

    permutations: amount of permutations for calculating p-value

    ### Returns
    The first float of the tuple is the statistic, a deterministic number that represents how
    extreme the difference between groups is. It is kind of impossible to interpret
    but its useful to compare different runs on the same data.
    The second float is the p-value, a permutation-based approximation of the
    probability that the null hypothesis should be accepted. Since the null hypothesis here
    is that there is no difference between groups, a low p-value (typically below 0.05)
    means that groups are likely different.
    """

ordinal_encoding_dtypes = np.str_ | np.int64 | np.int32 | np.int16
